﻿using System;
using System.Collections.Generic;
using System.Collections.Concurrent;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using Microsoft.AspNetCore.Mvc;
using Microsoft.Extensions.Configuration;
using DeviceModelLib;
using HubClientHandlerLib;

namespace BaseDeviceControllerLib
{
    [Route("api/[controller]")]
    public class BaseDeviceController : Controller
    {
        protected class Threshold
        {
            public float UpperLimit { get; private set; }
            public DeviceStatus Status { get; private set; }

            public Threshold(string s)
            {
                var ss = s.Split(',');
                switch (ss[0].ToLower())
                {
                    case "ok": Status = DeviceStatus.Ok; break;
                    case "warning": Status = DeviceStatus.Warning; break;
                    case "error": Status = DeviceStatus.Error; break;
                    default: Status = DeviceStatus.NoData; break;
                }

                UpperLimit = float.Parse(ss[1]);
            }
        }

        protected static HubClient hubClient;
        protected static List<Threshold> lstThreshold;
        protected static ConcurrentDictionary<int, DateTime> cdctLastUpdateTime = new ConcurrentDictionary<int, DateTime>();
        protected static Timer timerCheckIdle;
        protected static DeviceTypeInfo deviceTypeInfo;
        protected static bool isConfigured = false;

        protected BaseDeviceController(IConfiguration configuration)
        {
            if (!isConfigured)
                lock (typeof(BaseDeviceController))
                {
                    if (!isConfigured)
                    {
                        isConfigured = true;
                        ConfigureAndInit(configuration);
                    }
                }
        }

        protected static async void ConfigureAndInit(IConfiguration configuration)
        {
            var configSectionDashboard = configuration.GetSection("Dashboard");
            var configSectionDeviceTypeInfo = configuration.GetSection("DeviceTypeInfo");
            var configSectionTypeProcessingData = configuration.GetSection("TypeProcessingData");
            
            if (hubClient == null)
                hubClient = new HubClient();

            // This is to allow any order of starting applications 
            bool isConnected = false;
            while (!isConnected)
            {
                try
                {
                    await hubClient.StartAsync(configSectionDashboard.GetValue<string>("Uri"));
                    isConnected = true;
                }
                catch
                {
                    await Task.Delay(1000);
                }
            }

            var lstTempThreshold = new List<Threshold>();
            foreach (string s in configSectionTypeProcessingData.GetValue<string>("Thresholds").Split('|'))
            {
                string st = null;
                if (!string.IsNullOrEmpty(st = s.Replace(" ", string.Empty)))
                    lstTempThreshold.Add(new Threshold(st));
            }

            lstThreshold = lstTempThreshold.OrderBy(t => t.UpperLimit).ToList();

            int idleTimeIntervalInMs = configSectionTypeProcessingData.GetValue<int>("IdleTimeIntervalInMs");
            var idleTimeInterval = TimeSpan.FromMilliseconds(idleTimeIntervalInMs);

            deviceTypeInfo = new DeviceTypeInfo
                {
                    Id = configSectionDeviceTypeInfo.GetValue<int>("Id"),
                    Name = configSectionDeviceTypeInfo.GetValue<string>("Name"),
                    Min = configSectionDeviceTypeInfo.GetValue<float>("Min"),
                    Max = configSectionDeviceTypeInfo.GetValue<float>("Max"),
                    Units = configSectionDeviceTypeInfo.GetValue<string>("Units"),
                    Description = configSectionDeviceTypeInfo.GetValue<string>("Description"),
                    TypeSpecificInfo = configSectionDeviceTypeInfo.GetValue<string>("TypeSpecificInfo")
                };

            CallDashboardHub("RegisterDeviceTypeWithDashboard", deviceTypeInfo);
                             
            timerCheckIdle = new Timer(_ =>
                {
                    try
                    {
                        foreach (var deviceId in cdctLastUpdateTime.Keys)
                            if (DateTime.Now - cdctLastUpdateTime[deviceId] > idleTimeInterval)
                                CallDashboardHub("ForwardMeasurement", new Measurement { DeviceId = deviceId });
                    }
                    catch (Exception e)
                    {
                        // Writing to log
                    }
                },
                null, idleTimeIntervalInMs, idleTimeIntervalInMs);
        }

        [HttpGet]
        [Route("RegisterDevice/{deviceId}/{typeId}/{deviceName}/{description}/{deviceSpecificInfo}")]
        public virtual int RegisterDevice(int deviceId, int typeId, string deviceName, string description, string deviceSpecificInfo)
        {
            cdctLastUpdateTime[deviceId] = DateTime.Now;

            CallDashboardHub("RegisterDeviceWithDashboard",
                             new DeviceInfo
                             {
                                 Id = deviceId,
                                 TypeId = typeId,
                                 Name = deviceName,
                                 Description = description,
                                 DeviceSpecificInfo = deviceSpecificInfo
                             });

            return deviceId;
        }

        [HttpGet]
        [Route("SendMeasurement/{deviceId}/{data}")]
        public virtual ActionResult SendMeasurement(int deviceId, string data)
        {
            float value = float.NaN;
            float.TryParse(data, out value);
            value = GetTrueValue(value);
            var status = GetDeviceStatus(value);

            cdctLastUpdateTime[deviceId] = DateTime.Now;

            CallDashboardHub("ForwardMeasurement",
                new Measurement { DeviceId = deviceId, Status = status, Value = value });
            
            return Ok();
        }

        protected static void CallDashboardHub(string methodName, params object[] obs)
        {
            hubClient.InvokeRemoteAsync(methodName, obs);
        }

        protected virtual float GetTrueValue(float value)
        {
            //------------------------------------------------------------------------------------------------
            // Here (or in override method in derived classes) value transform may be implemented, if requied,
            // based on information specific to device's type and/or device instance.
            //------------------------------------------------------------------------------------------------
            return value;
        }

        protected virtual DeviceStatus GetDeviceStatus(float value)
        {
            var status = DeviceStatus.NoData;
            if (!float.IsNaN(value))
                foreach (var t in lstThreshold)
                    if (t.UpperLimit >= value)
                    {
                        status = t.Status;
                        break;
                    }

            return status;
        }
    }
}

