﻿using System;
using System.Collections.Generic;
using System.Threading;
using System.Threading.Tasks;
using Microsoft.AspNetCore.SignalR.Client;

namespace HubClientHandlerLib
{
    public class HubClient
    {
        public string Url { get; private set; }

        protected HubConnection hubConnection;
        protected CancellationTokenSource cancellationTokenSource = new CancellationTokenSource();

        public Dictionary<string, Action<object>> OnHandlers { protected get; set; } = new Dictionary<string, Action<object>>();

        public async Task StartAsync(string url)
        {
            Url = url;
            hubConnection = new HubConnectionBuilder().WithUrl(url).Build();

            foreach (var methodName in OnHandlers.Keys)
                hubConnection.On(methodName, OnHandlers[methodName]);

            try
            {
                await hubConnection.StartAsync();
            }
            catch (Exception e)
            {
                throw new Exception($"Hub connection on \"{url}\" had failed. ", e);
            }
        }

        public async void InvokeRemoteAsync(string methodName, params object[] args)
        {
            if (hubConnection == null)
                await Task.CompletedTask;

            var cancellationToken = cancellationTokenSource.Token;

            if (!cancellationToken.IsCancellationRequested && args != null && args.Length > 0)
            {
                try
                {
                    await hubConnection.InvokeAsync(methodName, args, cancellationToken);
                }
                catch (Exception e)
                {
                    throw new Exception($"Hub\"{Url}\" InvokeAsync() of method \"{methodName}()\" had failed. ", e);
                }
            }
            else
                await hubConnection.DisposeAsync();
        }

        public async void GetDataAndInvokeRemoteAsync(Func<object[]> func, string methodName, int delayInMs)
        {
            if (hubConnection == null)
                await Task.CompletedTask;

            var cancellationToken = cancellationTokenSource.Token;

            while (!cancellationToken.IsCancellationRequested)
            {
                await Task.Delay(delayInMs, cancellationToken);

                object[] obs = null;

                try
                {
                    obs = func();
                }
                catch (Exception e)
                {
                    // log....
                }

                // Finally send the value:
                if (obs != null && obs.Length > 0)
                    await hubConnection.InvokeAsync(methodName, obs, cancellationToken);
            }

            await hubConnection.DisposeAsync();
        }

        public void Cancel()
        {
            cancellationTokenSource.Cancel();
        }
    }
}

