import { Injectable } from '@angular/core';
import { HubConnection } from '@aspnet/signalr';
import { Observable } from 'rxjs';
import { Subject } from 'rxjs/Subject';

@Injectable()
export class CommunicationService {

    static comm: Communication;

    constructor() {
        //debugger;
        if (!CommunicationService.comm)
            CommunicationService.comm = new Communication();
    }
}

class Communication {
    BASE_URL = 'http://localhost:62631/hub/web';
    ARR_MAX_LENGTH = 10;

    devices: Array<DeviceInfo> = [];
    deviceTypes: Array<DeviceTypeInfo> = [];
    groups: Array<GroupInfo> = [];
    mapMeasurements: Map<number, Array<DeviceMeasurement>> = new Map<number, Array<DeviceMeasurement>>();

    private hubConnection: HubConnection;
    private isConnected: boolean = false;

    // Observable sources
    selectComponentSubject = new Subject<any>();
    streamSubject = new Subject<any>();
    receiveMeasurementSubject = new Subject<any>();
  
    constructor() {
        this.start();
    }
 
    async start() {
        this.hubConnection = new HubConnection(this.BASE_URL);

        // Handler on streaming push measurements
        this.hubConnection.on('registerDeviceWithWeb', result => this.onRegisterDevice(result.device));

        try {
            // Start connection
            await this.hubConnection.start();
            
            this.isConnected = true;
            
            await this.getInitInfo();
                
            // Handler to receive measurements
            await this.streamSubject.asObservable().subscribe(result => this.onReceiveMeasurements(result));

            await this.hubConnection.stream("StartStreaming").subscribe(this.streamSubject);
        }
        catch (err) { 
            console.error(err + ': Unable to establish connection with ' + `${this.BASE_URL}`);
        }
    }

    onReceiveMeasurements(result) {
        var measurements = result.measurements;

        if (!measurements)
            return;
        
        for (var i = 0; i < measurements.length; i++) {
            var measurement = measurements[i] as DeviceMeasurement;
            
            var arr = this.getMeasurementsByDeviceId(measurement.deviceId);
            if (!arr)   
                this.mapMeasurements.set(measurement.deviceId, new Array<DeviceMeasurement>());

            arr = this.getMeasurementsByDeviceId(measurement.deviceId);
            if (arr.length >= this.ARR_MAX_LENGTH)
                arr.shift();
            
            arr.push(measurement);
        }

        // Issue event for components
        this.receiveMeasurementSubject.next();
    }

    async getInitInfo() { 
        try {
            var result = await this.hubConnection.invoke('GetInitInfo');

            this.devices = result.devices;
            this.deviceTypes = result.deviceTypes;
            this.groups = result.groups;

            this.orderDevices();
        }
        catch(err) {
            console.error(err);
        }
    }
    
    onRegisterDevice(device) {
        if (!this.devices.find(d => d.id == device.id)) {
            this.devices.push(device);
            this.orderDevices();
        }
    }

    private orderDevices() {
        if (this.devices && this.devices.length > 0)
                this.devices.sort((a, b): number => a.id < b.id ? 0 : 1);
    }
  
    getMeasurementsByDeviceId(id: number): DeviceMeasurement[] {
        if (this.mapMeasurements)
            return this.mapMeasurements.get(id);
    }

    getLastMeasurementByDeviceId(id: number): DeviceMeasurement  {
        var measurements = this.getMeasurementsByDeviceId(id);
        return measurements && measurements.length > 0 ? measurements[measurements.length - 1] : undefined;
    }

    getDeviceById(id: number): DeviceInfo  {
        return this.devices.find(d => d.id == id);
    }

    getTypeByDeviceId(id: number): DeviceTypeInfo  {
        var device = this.getDeviceById(id);
        return device ? this.deviceTypes.find(t => t.id == device.typeId) : undefined;
    }
}

export class DeviceMeasurement {
    deviceId: number;
    value;
    timestamp;
    status: number;
}

class BaseInfo {
    id: number;
    name: string;
    description: string;
}

class DeviceInfo extends BaseInfo {
    typeId: number;
    groupId: number;
    deviceSpecificInfo: string;
}

class DeviceTypeInfo extends BaseInfo {
    units: string;
    min;
    max;
    typeSpecificInfo: string;
}

class GroupInfo extends BaseInfo { }
